#===============================================================================#
#                       GoBright Interact Serial Flasher Helper 
#===============================================================================#
#
#                               Copyright © GoBright
#
# This script and any accompanying firmware files are the exclusive property of GoBright.
# Unauthorized use, reproduction, or distribution of this script or firmware files is strictly prohibited.
#
# This tool is intended for use only under the guidance of GoBright support.
# Please handle this tool with care, as improper use may result in unintended consequences.
#
# Note: This script uses esptool, which is licensed under the GNU General Public License v2.
# For more details on this license, visit: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
#
# =============================================================================== #

$VERSION = "V1.5"

# =============================================================================== #
# Function declarations
# =============================================================================== #

# Function to read ASCII string from a binary file
function Get-AsciiStringFromBinary {
    param (
        [Parameter(Mandatory = $true)]
        [string]$filePath,
        [Parameter(Mandatory = $true)]
        [int]$startAddress,
        [Parameter(Mandatory = $true)]
        [int]$maxLength
    )

    try {
        $bytes = [System.IO.File]::ReadAllBytes($filePath)
        if ($bytes.Length -lt $startAddress + $maxLength) {
            return $null
        }

        $stringBytes = @()
        for ($i = $startAddress; $i -lt ($startAddress + $maxLength); $i++) {
            if ($bytes[$i] -eq 0x00) {
                break
            }
            $stringBytes += $bytes[$i]
        }

        if ($stringBytes.Count -eq 0) {
            throw "No valid string found at the specified address."
        }

        return [System.Text.Encoding]::ASCII.GetString($stringBytes)
    }
    catch {
        return $null
    }
}

# =============================================================================== #

# Function to check if a port is an ESP device
function Test-EspPort {
    param (
        [Parameter(Mandatory = $true)]
        [string]$port
    )
    try {
        $result = & "$PSScriptRoot\InteractFlashTools\esptool.exe" --port $port --connect-attempts 4 read_mac 2>&1
        # Write-Host "$result" -ForegroundColor Yellow
        return ($result -like "*Chip is ESP32-S3*" -and $result -like "*MAC:*")
    }
    catch {
        return $false
    }
}

# =============================================================================== #

# Function to check if a port is an ESP device
function Select-PortFromList {
    param (
        [Parameter(Mandatory = $true)]
        [Array]$ports,
        [Parameter(Mandatory = $true)]
        [bool]$printList
    )
    
    while ($true) {
        if($printList) {
            foreach ($port in $ports) {
                Write-Host " - $($port)"
            }
        }
        $selectedPortNumber = Read-Host -Prompt 'Enter serial port number'
        if ($selectedPortNumber.ToLower() -eq "q") {
            Clear-Host
            exit
        }
        $selectedPort = "COM$selectedPortNumber"
        foreach ($port in $ports) {
            if ($port -eq $selectedPort) {
                return $port
            }
        }
        Write-Host "Invalid port number. Please enter a valid serial port number from the list." -ForegroundColor Yellow
    }
}

# =============================================================================== #
# Start of script logic
# =============================================================================== #

# Check if esptool.exe exists
if (-Not (Test-Path "$PSScriptRoot\InteractFlashTools\esptool.exe")) { Write-Warning "Could not find esptool.exe in the local InteractFlashTools folder, This can be downloaded from https://github.com/espressif/esptool/releases"; exit }
$global:fwfile = $null

# default values
$global:flashEntireDevice = $false
$global:eraseFlashFirst = $false
$global:scanForFiles = $true
$global:autoSelect = $true
$global:blacklist = @()

# =============================================================================== #
# start loop to do executions
# =============================================================================== #
:startOfLoop for (; ; ) {

    # =============================================================================== #
    # Scan for files on first call and when a rescan is triggered
    # =============================================================================== #
    if ($global:scanForFiles) {
        # Check if firmware folder exists
        if (!(Test-Path "$PSScriptRoot\InteractFirmwareFiles\" -PathType Container)) {
            Write-Warning "Could not find Interact firmware folder 'InteractFirmwareFiles', exiting..."
            exit
        }

        # Check if the additional firmware files exist.
        $global:allAdditionalFilesExist = (Test-Path "$PSScriptRoot\InteractFirmwareFiles\interact-bootloader.bin") -And 
                                          (Test-Path "$PSScriptRoot\InteractFirmwareFiles\interact-partition-table.bin") -And 
                                          (Test-Path "$PSScriptRoot\InteractFirmwareFiles\interact-ota_data_initial.bin") -And 
                                          (Test-Path "$PSScriptRoot\InteractFirmwareFiles\interact-phy_init_data.bin")

        # Retrieve firmware file using regex
        $fwfiles = Get-ChildItem -Path "$PSScriptRoot\InteractFirmwareFiles" | Where-Object { $_.Name -match '^interact-\d+\.\d+\.\d+.*\.bin$' }

        if (($fwfiles | Measure-Object).Count -eq 0) {
            Write-Warning "Could not find suitable Interact fw file, exiting..."
            exit
        }

        $global:moreThenOneFirmwareFile = $(($fwfiles | Measure-Object).Count -gt 1)

        $global:scanForFiles = $false
    }

    Clear-Host
    # =============================================================================== #
    # Legal and Usage Information
    # =============================================================================== #
    Write-Host "===============================================================================" -ForegroundColor Cyan
    Write-Host "                  GoBright Interact Serial Flasher Helper " -ForegroundColor Yellow -NoNewline
    Write-Host "$VERSION" -ForegroundColor Cyan
    Write-Host "===============================================================================" -ForegroundColor Cyan
    Write-Host "`n                             Copyright " -NoNewline -ForegroundColor Blue
    Write-Host "$([char]0x00A9) " -NoNewline -ForegroundColor Cyan
    Write-Host "GoBright`n" -ForegroundColor Blue
    Write-Host "This script and any accompanying firmware files are the exclusive property of GoBright."
    Write-Host "Unauthorized use, reproduction, or distribution of this script or firmware files is strictly prohibited.`n"
    Write-Host "This tool is intended for use only under the guidance of GoBright support."
    Write-Host "Please handle this tool with care, as improper use may result in unintended consequences.`n"
    Write-Host "Note: This script uses esptool, which is licensed under the GNU General Public License v2."
    Write-Host "For more details on this license, visit: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html`n"
    Write-Host "===============================================================================" -ForegroundColor Cyan

    # =============================================================================== #
    # Select firmware file if not yet selected
    # =============================================================================== #
    if (!$global:fwfile) {
        if ($global:moreThenOneFirmwareFile) {
            Write-Host "Multiple ($($fwfiles.Count)) firmware files detected:"
            for ($i = 0; $i -lt $fwfiles.Count; $i++) {
                Write-Host "[" -NoNewline
                Write-Host "$("{0,3}" -f $i)" -ForegroundColor Green -NoNewline
                Write-Host "] " -NoNewline
                Write-Host "$($fwfiles[$i].Name)" -ForegroundColor Yellow
            }

            while ($true) {
                $selectedIndex = Read-Host "Enter the number of the firmware file to use or `"R`" to rescan"
                if ($selectedIndex -match '^\d+$' -and [int]$selectedIndex -ge 0 -and [int]$selectedIndex -lt $fwfiles.Count) {
                    $global:fwfile = $fwfiles[$selectedIndex]
                    break
                }
                elseif ($selectedIndex.ToLower() -eq "r") {
                    $global:scanForFiles = $true
                    continue startOfLoop
                }
                elseif ($selectedIndex.ToLower() -eq "q") {
                    Clear-Host
                    exit
                }
                else {
                    Write-Host "Invalid selection. Please enter a valid number from the list or `"R`" to rescan" -ForegroundColor Yellow
                }
            }
            Write-Host "===============================================================================" -ForegroundColor Cyan
        }
        else {
            $global:fwfile = $fwfiles[0]
        }
    }

    Write-Host "Selected firmware File: " -NoNewline
    Write-Host "$($global:fwfile.Name)" -ForegroundColor Yellow
    Write-Host "-------------------------------------------------------------------------------" -ForegroundColor Cyan

    $firmwareVersion = Get-AsciiStringFromBinary -filePath $global:fwfile.FullName -startAddress 0x30 -maxLength 50
    $espIdfVersion = Get-AsciiStringFromBinary -filePath $global:fwfile.FullName -startAddress 0x90 -maxLength 50

    if ($firmwareVersion -And $espIdfVersion) {
        Write-Host "Detected firmware version: " -NoNewline
        Write-Host "$firmwareVersion" -ForegroundColor Green -NoNewline
        Write-Host " (IDF version: " -NoNewline
        Write-Host "$espIdfVersion" -ForegroundColor Green -NoNewline
        Write-Host ")"
    }
    elseif ($firmwareVersion -And !$espIdfVersion) {
        Write-Host "Detected firmware version: " -NoNewline
        Write-Host "$firmwareVersion" -ForegroundColor Green
    }

    Write-Host "===============================================================================" -ForegroundColor Cyan

    $selectableLetterColor = "DarkCyan"

    # =============================================================================== #
    # Flash mode (entire flash or applicition only)
    # =============================================================================== #
    if ($global:allAdditionalFilesExist){
        Write-Host "Press ['" -NoNewline
        Write-Host "F"  -NoNewline -ForegroundColor $selectableLetterColor
        Write-Host "'] to change the flashing mode, " -NoNewline
        Write-Host "currently in " -NoNewline
        Write-Host "$(if ($global:flashEntireDevice) {"'flash entire device'"} else {"'flash application only'"})" -ForegroundColor Yellow -NoNewline
        Write-Host " mode "
    }
    
    # =============================================================================== #
    # Erasure mode, erase entire flash before flashing
    # =============================================================================== #
    if ($global:allAdditionalFilesExist){
        Write-Host "Press ['" -NoNewline -ForegroundColor $(if ($global:flashEntireDevice) { "Gray" } else { "DarkGray" })
        Write-Host "E" -NoNewline -ForegroundColor $selectableLetterColor
        Write-Host "'] to toggle erasing flash before flashing, " -NoNewline -ForegroundColor $(if ($global:flashEntireDevice) { "Gray" } else { "DarkGray" })
        Write-Host "currently " -NoNewline
        Write-Host "$(if($global:eraseFlashFirst) {"'Enabled'"} else {"'Disabled'"})" -ForegroundColor Yellow
    }
    # =============================================================================== #
    # Firmware File selection
    # =============================================================================== #
    Write-Host "Press ['" -NoNewline -ForegroundColor $(if ($global:moreThenOneFirmwareFile) { "Gray" } else { "DarkGray" }) 
    Write-Host "S" -NoNewline -ForegroundColor $selectableLetterColor
    Write-Host "'] to Select another Firmware File, " -ForegroundColor $(if ($global:moreThenOneFirmwareFile) { "Gray" } else { "DarkGray" }) -NoNewline
    Write-Host "$(if(($fwfiles | Measure-Object).Count -gt 1) {"($(($fwfiles | Measure-Object).Count)) files found"})" -ForegroundColor Yellow
    
    # =============================================================================== #
    # Rescan for files
    # =============================================================================== #
    Write-Host "Press ['" -NoNewline  -ForegroundColor Gray
    Write-Host "R" -NoNewline -ForegroundColor $selectableLetterColor
    Write-Host "'] to rescan for files " -ForegroundColor Gray

    # =============================================================================== #
    # Com port detection method
    # =============================================================================== #
    Write-Host "Press ['" -ForegroundColor Gray -NoNewline
    Write-Host "D" -NoNewline -ForegroundColor $selectableLetterColor
    Write-Host "'] to change the port detection method, currently in " -NoNewline
    Write-Host "$(if ($global:autoSelect) {"'Autodetect'"} else {"'Manual Select'"})" -ForegroundColor Yellow -NoNewline
    Write-Host " mode" -ForegroundColor Gray -NoNewline
    Write-Host " $(if ($global:autoSelect) {"(Experimental)"})" -ForegroundColor Magenta
    
    # =============================================================================== #
    # Clear Blacklist option
    # =============================================================================== #
    if ($global:autoSelect) {
        Write-Host "Press ['" -NoNewline -ForegroundColor $(if ($global:blacklist.Length -eq 0) { "DarkGray" } else { "Gray" })
        Write-Host "C" -NoNewline -ForegroundColor $selectableLetterColor
        Write-Host "'] to clear the com port autodetect blacklist"  -NoNewline -ForegroundColor $(if ($global:blacklist.Length -eq 0) { "DarkGray" } else { "Gray" })
        Write-Host " $(if ($global:blacklist.Length -gt 0) { "({0})" -f ($global:blacklist -join ", ") })" -ForegroundColor Yellow
    }
    
    # =============================================================================== #
    # Exit script
    # =============================================================================== #
    Write-Host "Press ['" -NoNewline -ForegroundColor Red
    Write-Host "Q" -NoNewline -ForegroundColor $selectableLetterColor
    Write-Host "'] to exit" -ForegroundColor Red

    # =============================================================================== #
    # Continue operation
    # =============================================================================== #
    Write-Host "Press ['" -NoNewline -ForegroundColor Green
    Write-Host "ENTER" -NoNewline -ForegroundColor $selectableLetterColor
    Write-Host "'] to start the process " -NoNewline -ForegroundColor Green
    Write-Host "(Make sure the Interact is connected and powered on)" -ForegroundColor yellow

    Write-Host "-------------------------------------------------------------------------------" -ForegroundColor Cyan

    # =============================================================================== #
    # Input handler switch
    # =============================================================================== #
    Write-Host "Selection: " -NoNewline
    $keyInfo = [System.Console]::ReadKey($true)
    
    if (!($keyInfo.Key -eq [ConsoleKey]::Enter)) {
        Write-Host ""
        switch ($keyInfo.KeyChar.ToString().ToLower()) {
            # Flash mode (entire flash or applicition only)
            "f" {
                $global:flashEntireDevice = $(!$global:flashEntireDevice -And $global:allAdditionalFilesExist)
                if (!$global:flashEntireDevice -And $global:eraseFlashFirst) {
                    $global:eraseFlashFirst = $false
                }
                continue startOfLoop
            }
            # Erasure mode, erase entire flash before flashing
            "e" {
                $global:eraseFlashFirst = $(!$global:eraseFlashFirst -And $global:allAdditionalFilesExist -And $global:flashEntireDevice)
                continue startOfLoop
            }
            # Firmware File selection
            "s" {
                if ($global:moreThenOneFirmwareFile) {
                    $global:fwfile = $false
                }
                continue startOfLoop
            }
            # Rescan for files
            "r" {
                $global:scanForFiles = $true
                continue startOfLoop
            }
            # Com port detection method
            "d" {
                $global:autoSelect = -not $global:autoSelect
                continue startOfLoop
            }
            # Clear Blacklist optionf
            "c" {
                if ($global:autoSelect) {
                    $global:blacklist = @()
                }
                continue startOfLoop
            }
            # Exit script
            "q" {
                Clear-Host
                exit
            }
            default {
                continue startOfLoop
            }
        }
    } else {
        Write-Host "Starting process..." -ForegroundColor Green
    }
    # =============================================================================== #
    # Select serial port
    # =============================================================================== #

    $serialports = Get-CimInstance -Class Win32_SerialPort | Select-Object Name, Description, DeviceID
    if (($serialports | Measure-Object).Count -eq 0) {
        if ($(Read-Host "No serial port found, hit [ENTER] to start over") -ieq "q") {
            Clear-Host
            exit
        }
        else {
            continue
        }
    }

    $serialPortIDs = @()

    Write-Host "Found serial ports:"
    Foreach ($serialport in $serialports) {
        $serialPortIDs += $($serialport.DeviceID) 
        Write-Host " - $($serialport.DeviceID)"
    }

    $comportDeviceId = $null

    if ($global:autoSelect) {
        # Auto-select mode
        $espPortIDs = @()
        foreach ($serialportID in $serialPortIDs) {
            if ($global:blacklist -notcontains $serialportID) {
                Write-Host "Checking port $serialportID... " -NoNewline
                if (Test-EspPort -port $serialportID) {
                    $espPortIDs += $serialportID
                    Write-Host "Is ESP device"
                }
                else {
                    $global:blacklist += $serialportID
                    Write-Host "Is Not ESP device, adding to blacklist"
                }
            }
            else {
                Write-Host "$serialportID was blacklisted in previous attempts"
            }
        }

        if ($espPortIDs.Count -eq 0) {
            if ($(Read-Host "No ESP device found, hit [ENTER] to start over") -ieq "q") {
                Clear-Host
                exit
            }
            else {
                continue startOfLoop
            }
        }
        elseif ($espPortIDs.Count -eq 1) {
            $comportDeviceId = $espPortIDs[0]
        }
        else {
            Write-Output "Multiple ESP devices found:"
            $comportDeviceId = Select-PortFromList -ports $espPortIDs -printList $true
        }
    }
    else {
        # Manual mode
        if ($serialPortIDs.Count -eq 1) {
            $comportDeviceId = $serialPortIDs[0]
        }
        else {
            $comportDeviceId = Select-PortFromList -ports $serialPortIDs -printList $false
        }
    }

    Write-Host "-------------------------------------------------------------------------------" -ForegroundColor Cyan
    Write-Host "Will use serial port: " -NoNewline
    Write-Host "$comportDeviceId" -ForegroundColor Green
    Write-Host "-------------------------------------------------------------------------------" -ForegroundColor Cyan

    # =============================================================================== #
    # execute flashing
    # =============================================================================== #

    # Define the common part of the command
    $commonCommand = "& `"$PSScriptRoot\InteractFlashTools\esptool.exe`" --chip esp32s3 --port $comportDeviceId --baud 460800 --before default_reset --after hard_reset write_flash --flash_mode dio --flash_freq 80m --flash_size 16MB --no-compress"

    # Define the different sets of arguments
    $AppOnly = "0x120000 `"$($global:fwfile.FullName)`" 0x890000 `"$($global:fwfile.FullName)`""
    $FullDevice = "0x0 InteractSystemFiles\interact-bootloader.bin 0xb000 InteractSystemFiles\interact-partition-table.bin 0xc000 InteractSystemFiles\interact-ota_data_initial.bin 0xe000 InteractSystemFiles\interact-phy_init_data.bin"

    # Execute the final command
    Invoke-Expression "$commonCommand $(if($global:eraseFlashFirst) {"--erase-all"}) $AppOnly $(if($global:flashEntireDevice -And $global:allAdditionalFilesExist) {$FullDevice})"

    # =============================================================================== #
    # Finshed
    # =============================================================================== #
    Write-Host "===============================================================================" -ForegroundColor Cyan
    if ($(Read-Host "Finished, hit [ENTER] to start over") -ieq "q") {
        Clear-Host
        exit
    }
}